/* nac - nac.c (21092004)
 * Copyright (C) 2004 Craig McPartland

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#if TIME_WITH_SYS_TIME
# include <sys/time.h>          /* select? */
# include <time.h>              /* localtime, time */
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#if STDC_HEADERS
# include <string.h>            /* strcmp, strrchr */
#else
# if !HAVE_STRCHR
#  define strrchr rindex
# endif
#endif
#include <sys/select.h>         /* select? (POSIX) */
#include <sys/types.h>          /* select? */
#include <unistd.h>             /* select? */
#include <stdio.h>              /* fprintf, sprintf */
#include <stdlib.h>             /* atoi, exit, strtol */
#include <math.h>               /* cos, sin */
#include <ctype.h>              /* isxdigit */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>

#define BUTTON_1 0x100

#define CLOCK_WIDTH 100
#define CLOCK_HEIGHT 100

char *title;
Display *display;

void set_colour(char *, unsigned long *, char *);
int is_hex(char *);
void draw_digits(Window, Pixmap, int, char *, int);
void draw_hands(Window, GC, int, int, int);
void usage(char *);
XFontStruct *getfont(char *);

int main(int argc, char *argv[])
{
  Atom delete_window, protocols;
  Bool shape_ext;
  char *font = NULL;
  char *back_colour = NULL, *digit_colour = NULL, *hand_colour = NULL;
  char *display_name = NULL, *image = NULL, *option;
  GC gc;
  int default_depth, default_screen, display_width, display_height;
  int h_len = 25, m_len = 40, s_len = 45, d_len = 45;
  int arg, nodigits = 0, xfd, geomask = 0, evt_base, err_base, running = 1;
  int cx = 0, cy = 0, imagex = 0, imagey = 0, top = 0, wm = 0;
  Pixmap backpix, pixmap;
  unsigned int width, height, nc;
  unsigned long back_pixel, digit_pixel, hand_pixel, black_pixel,
      white_pixel;
  Window window, root_window, saver = None, root, parent, *child;
  XGCValues values;
  XpmAttributes xpm_attr;
  XSizeHints size_hints;
  XWMHints wm_hints;
  XClassHint class_hint;
  XTextProperty text_property;
  XWindowAttributes wa;

  /* Get name of this program. */
  if ((title = strrchr(argv[0], '/')) == NULL)
    title = argv[0];
  else
    title++;

  /* Get program arguments. */
  for (arg = 1; arg < argc; arg++) {
    if (!strcmp(argv[arg], "-face")) {
      if (argv[++arg] == NULL)
        usage(title);
      back_colour = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-digits")) {
      if (argv[++arg] == NULL)
        usage(title);
      digit_colour = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-display")) {
      if (argv[++arg] == NULL)
        usage(title);
      display_name = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-image")) {
      if (argv[++arg] == NULL)
        usage(title);
      image = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-imagex")) {
      if (argv[++arg] == NULL)
        usage(title);
      imagex = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-imagey")) {
      if (argv[++arg] == NULL)
        usage(title);
      imagey = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-geometry")) {
      if (argv[++arg] == NULL)
        usage(title);
      geomask = XParseGeometry(argv[arg], &cx, &cy, &width, &height);
      continue;
    }
    if (!strcmp(argv[arg], "-hands")) {
      if (argv[++arg] == NULL)
        usage(title);
      hand_colour = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-font")) {
      if (argv[++arg] == NULL)
        usage(title);
      font = argv[arg];
      continue;
    }
    if (!strcmp(argv[arg], "-hlen")) {
      if (argv[++arg] == NULL)
        usage(title);
      h_len = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-mlen")) {
      if (argv[++arg] == NULL)
        usage(title);
      m_len = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-slen")) {
      if (argv[++arg] == NULL)
        usage(title);
      s_len = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-drad")) {
      if (argv[++arg] == NULL)
        usage(title);
      d_len = atoi(argv[arg]);
      continue;
    }
    if (!strcmp(argv[arg], "-nodigits")) {
      nodigits = 1;
      continue;
    }
    if (!strcmp(argv[arg], "-top")) {
      top = 1;
      continue;
    }
    if (!strcmp(argv[arg], "-wm")) {
      wm = 1;
      continue;
    }
    usage(title);
  }

  /* Open display. */
  if ((display = XOpenDisplay(display_name)) == NULL) {
    fprintf(stderr, "%s: XOpenDisplay failed.\n", title);
    exit(EXIT_FAILURE);
  }

  /* Get default options. */
  if (back_colour == NULL)
    if ((option = XGetDefault(display, title, "Background")) != NULL) {
      back_colour = option;
    }

  if (digit_colour == NULL) {
    if ((option = XGetDefault(display, title, "digits")) != NULL)
      digit_colour = option;
  }

  if (image == NULL) {
    if ((option = XGetDefault(display, title, "image")) != NULL)
      image = option;
  }

  if (!imagex) {
    if ((option = XGetDefault(display, title, "imagex")) != NULL)
      imagex = atoi(option);
  }

  if (!imagey) {
    if ((option = XGetDefault(display, title, "imagey")) != NULL)
      imagey = atoi(option);
  }

  if (hand_colour == NULL) {
    if ((option = XGetDefault(display, title, "hands")) != NULL)
      hand_colour = option;
  }

  default_screen = DefaultScreen(display);
  display_width = XDisplayWidth(display, default_screen);
  display_height = XDisplayHeight(display, default_screen);

  if (!(geomask & XValue) && !(geomask & XNegative)) {
    if ((option = XGetDefault(display, title, "x")) != NULL) {
      if ((cx = atoi(option)) < 0)
        cx += display_width - CLOCK_WIDTH;
    }
  }

  if (!(geomask & YValue) && !(geomask & YNegative)) {
    if ((option = XGetDefault(display, title, "y")) != NULL) {
      if ((cy = atoi(option)) < 0)
        cy += display_height - CLOCK_HEIGHT;
    }
  }

  /* Set window position. */
  if (geomask & XNegative)
    cx += display_width - CLOCK_WIDTH;
  if (geomask & YNegative)
    cy += display_height - CLOCK_HEIGHT;

  /* Create Window. */
  root_window = RootWindow(display, default_screen);
  black_pixel = BlackPixel(display, default_screen);
  white_pixel = WhitePixel(display, default_screen);
  window = XCreateSimpleWindow(display, root_window, cx, cy, CLOCK_WIDTH,
                               CLOCK_HEIGHT, 1, black_pixel, white_pixel);

  /* Shape Window. */
  shape_ext = XShapeQueryExtension(display, &evt_base, &err_base);
  if (shape_ext) {
    pixmap = XCreatePixmap(display, window, CLOCK_WIDTH, CLOCK_HEIGHT, 1);
    gc = XCreateGC(display, pixmap, 0, &values);
    XSetForeground(display, gc, 0);
    XFillRectangle(display, pixmap, gc, 0, 0, CLOCK_WIDTH, CLOCK_HEIGHT);
    XSetForeground(display, gc, 1);
    XFillArc(display, pixmap, gc, 0, 0, CLOCK_WIDTH, CLOCK_HEIGHT, 0,
             23040);
    XShapeCombineMask(display, window, ShapeBounding, 0, 0, pixmap,
                      ShapeSet);
    XFreeGC(display, gc);
    XFreePixmap(display, pixmap);
  }

  /* Set colour of digits. */
  digit_pixel = white_pixel;
  set_colour(digit_colour, &digit_pixel, title);

  /* Set colour of hands. */
  hand_pixel = white_pixel;
  set_colour(hand_colour, &hand_pixel, title);

  /* Set colour of background. */
  back_pixel = black_pixel;
  set_colour(back_colour, &back_pixel, title);

  /* Create background pixmap. */
  default_depth = DefaultDepth(display, default_screen);
  backpix = XCreatePixmap(display, window, CLOCK_WIDTH, CLOCK_HEIGHT,
                          default_depth);
  gc = XCreateGC(display, backpix, 0, &values);
  XSetForeground(display, gc, back_pixel);
  XFillRectangle(display, backpix, gc, 0, 0, CLOCK_WIDTH, CLOCK_HEIGHT);
  XFreeGC(display, gc);

  /* Put image on clock face. */
  xpm_attr.valuemask = 0;
  if (image != NULL) {
    Pixmap bitmap;
    int stat = XpmReadFileToPixmap(display, window, image,
                                   &pixmap, &bitmap, &xpm_attr);
    if (stat == XpmSuccess) {
      values.clip_mask = bitmap;
      values.clip_x_origin = imagex + CLOCK_WIDTH / 2 - xpm_attr.width / 2;
      values.clip_y_origin =
          imagey + CLOCK_HEIGHT / 2 - xpm_attr.height / 2;
      gc = XCreateGC(display, backpix,
                     GCClipMask | GCClipXOrigin | GCClipYOrigin, &values);
      XCopyArea(display, pixmap, backpix, gc, 0, 0, xpm_attr.width,
                xpm_attr.height, values.clip_x_origin,
                values.clip_y_origin);
      XFreeGC(display, gc);
      if (bitmap != None)
        XFreePixmap(display, bitmap);
      XFreePixmap(display, pixmap);
      XpmFreeAttributes(&xpm_attr);
    } else {
      fprintf(stderr, "%s: Failed to load %s: ", title, image);
      fprintf(stderr, "%s.\n", XpmGetErrorString(stat));
    }
  }

  if (!nodigits)
    draw_digits(window, backpix, d_len, font, digit_pixel);

  XSetWindowBackgroundPixmap(display, window, backpix);
  XFreePixmap(display, backpix);

  /* Create window GC. */
  values.foreground = hand_pixel;
  values.background = back_pixel;
  values.line_width = 0;
  gc = XCreateGC(display, window,
                 GCForeground | GCBackground | GCLineWidth, &values);

  if (!wm) {
    /* Tell window manager to leave window alone. */
    XSetWindowAttributes window_attributes;
    window_attributes.override_redirect = True;
    XChangeWindowAttributes(display, window, CWOverrideRedirect,
                            &window_attributes);
  }

  /* Want to receive client delete window message from window manager. */
  protocols = XInternAtom(display, "WM_PROTOCOLS", True);
  delete_window = XInternAtom(display, "WM_DELETE_WINDOW", True);
  XSetWMProtocols(display, window, &delete_window, 1);

  /* Give hints about window to the window manager. */
  wm_hints.flags = StateHint;
  wm_hints.initial_state = NormalState;

  /* Give hints about application name and class for resource database. */
  class_hint.res_name = title;
  class_hint.res_class = "game";

  /* Give hints about the window size. */
  size_hints.min_width = CLOCK_WIDTH;
  size_hints.max_width = CLOCK_WIDTH;
  size_hints.min_height = CLOCK_HEIGHT;
  size_hints.max_height = CLOCK_HEIGHT;
  size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;

  XSetWMProperties(display, window, NULL, NULL, argv, argc,
                   &size_hints, &wm_hints, &class_hint);

  /* Tell window manager what the window wants to be called. */
  if (!XStringListToTextProperty(&title, 1, &text_property))
    fprintf(stderr, "%s: XSetWMName failed.\n", title);
  else
    XSetWMName(display, window, &text_property);

  /* Request X server report certain events. */
  XSelectInput(display, window, ExposureMask | ButtonPressMask |
               PointerMotionMask | PropertyChangeMask |
               VisibilityChangeMask);

  /* Get window ID of xscreensaver (if running) for use later. */
  if (XQueryTree(display, root_window, &root, &parent, &child, &nc)) {
    char *name;
    int i;
    for (i = 0; i < nc && saver == None; i++) {
      if (XFetchName(display, child[i], &name)) {
        if (name != NULL) {
          if (!strcmp(name, "screensaver"))
            saver = child[i];
          XFree(name);
        }
      }
    }
    if (child != NULL)
      XFree(child);
  }

  /* Show window. */
  XMapWindow(display, window);

  /* Get display connection file descriptor. */
  xfd = ConnectionNumber(display);

  /* Event loop. */
  while (running) {
    fd_set rfds;
    int pending;
    struct timeval timeout;

    FD_ZERO(&rfds);
    FD_SET(xfd, &rfds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    /* Wait up to one second for window events, then draw hands. */
    if (!XPending(display))
      if (!select(xfd + 1, &rfds, NULL, NULL, &timeout)) {
        draw_hands(window, gc, h_len, m_len, s_len);
        continue;
      }

    /* Handle window events. */
    pending = XPending(display);

    while (pending--) {
      XEvent event;

      XNextEvent(display, &event);
      switch (event.type) {
        int x, y;
      case ButtonPress:        /* Involved in moving unmanaged window. */
        x = event.xbutton.x_root;
        y = event.xbutton.y_root;
        if (event.xbutton.button == 1) {        /* Get info. for window drag. */
          XRaiseWindow(display, window);
          XGetWindowAttributes(display, window, &wa);
          cx = wa.x - x;
          cy = wa.y - y;
        } else if (event.xbutton.button == 2)   /* Second button to exit. */
          running = 0;
        break;
      case ClientMessage:      /* Check if wm wants to delete window. */
        if (event.xclient.message_type == protocols &&
            event.xclient.data.l[0] == delete_window)
          running = 0;
        break;
      case MotionNotify:       /* Involved in moving unmanaged window. */
        x = event.xmotion.x_root;
        y = event.xmotion.y_root;
        if (event.xmotion.state & BUTTON_1 && !wm)      /* Window drag. */
          XMoveWindow(display, window, x + cx, y + cy);
        break;
      case VisibilityNotify:   /* Move to top if always on top wanted. */
        if (top
            && (event.xvisibility.state == VisibilityPartiallyObscured
                || event.xvisibility.state == VisibilityFullyObscured)) {
          if (saver != None) {  /* Don't put on top of screensaver window. */
            XGetWindowAttributes(display, saver, &wa);
            if (wa.map_state == IsUnmapped)
              XRaiseWindow(display, window);
          } else                /* No screensaver window. */
            XRaiseWindow(display, window);
        }
        break;
      default:
        break;
      }
    }
    /* Draw hands after events or there will be a delay before redraw. */
    draw_hands(window, gc, h_len, m_len, s_len);
  }

  XFreeGC(display, gc);
  XDestroyWindow(display, window);
  XCloseDisplay(display);

  return 0;
}

void set_colour(char *colour, unsigned long *value, char *title)
{
  if (colour != NULL) {
    Colormap colourmap = DefaultColormap(display, DefaultScreen(display));
    XColor screen_def, exact_def;

    if (XAllocNamedColor
        (display, colourmap, colour, &screen_def, &exact_def))
      *value = screen_def.pixel;
    else if (is_hex(colour)) {
      screen_def.pixel = strtol(colour, NULL, 16);
      screen_def.red = screen_def.pixel >> 16 & 0xff;
      screen_def.green = screen_def.pixel >> 8 & 0xff;
      screen_def.blue = screen_def.pixel & 0xff;
      screen_def.red *= screen_def.red;
      screen_def.green *= screen_def.green;
      screen_def.blue *= screen_def.blue;
      if (XAllocColor(display, colourmap, &screen_def))
        *value = screen_def.pixel;
      else
        fprintf(stderr, "%s: Couldn't set colour %s.\n", title, colour);
    } else
      fprintf(stderr, "%s: Couldn't set colour %s.\n", title, colour);
  }
}

int is_hex(char *hex)
{
  int i;

  for (i = 0; i < strlen(hex); i++)
    if (!isxdigit(hex[i]) && hex[i] != 'x')
      return 0;

  return 1;
}

void draw_digits(Window w, Pixmap pixmap, int radius, char *font,
                 int colour)
{
  char hour[3];
  float angle;
  GC gc;
  int i, x, y;
  XFontStruct *xfs;
  XGCValues values;

  xfs = getfont(font);
  if (xfs == NULL) {
    fprintf(stderr, "%s: Couldn't get any fonts!\n", title);
    return;
  }
  values.foreground = colour;
  values.font = xfs->fid;
  gc = XCreateGC(display, pixmap, GCForeground | GCBackground | GCFont,
                 &values);
  for (i = 0; i < 12; i++) {
    angle = (i - 2) * -0.5235;
    x = 48 + radius * cos(angle);
    y = 54 - radius * sin(angle);
    sprintf(hour, "%d", i + 1);
    XDrawString(display, pixmap, gc, x, y, hour, strlen(hour));
  }

  XFreeGC(display, gc);
}

XFontStruct *getfont(char *fnt)
{
  char *font[] = { "-misc-fixed-*-*-*-*-12-*-*-*-*-*-*-*",
    "-*-fixed-*-*-*-*-12-*-*-*-*-*-*-*", "-*-*-*-*-*-*-12-*-*-*-*-*-*-*",
    "*-*-*-*-*-*-10-*-*-*-*-*-*-*", "*-*-*-*-*-*-*-*-*-*-*-*-*-*", NULL
  };
  int i = 0;
  XFontStruct *xfs = NULL;

  if (fnt != NULL) {
    xfs = XLoadQueryFont(display, fnt);
    if (xfs != NULL)
      return xfs;
    else
      fprintf(stderr, "%s: Couldn't get font %s.\n", title, fnt);
  }

  while (xfs == NULL && font[i] != NULL)
    xfs = XLoadQueryFont(display, font[i++]);

  return xfs;
}

void draw_hands(Window w, GC gc, int h_len, int m_len, int s_len)
{
  double angle;
  int min, x, y;
  struct tm *now;
  time_t t;

  XClearWindow(display, w);

  t = time(NULL);
  now = localtime(&t);

  min = now->tm_min;

  /* Draw hour hand. */
  angle = (((now->tm_hour % 12) + min / 60.0) - 3) * -0.5235;
  x = h_len * cos(angle);
  y = h_len * sin(angle);
  XDrawLine(display, w, gc, 50, 50, 50 + x, 50 - y);

  /* Draw minute hand. */
  angle = (min - 15) * -0.1047;
  x = m_len * cos(angle);
  y = m_len * sin(angle);
  XDrawLine(display, w, gc, 50, 50, 50 + x, 50 - y);

  /* Draw second hand. */
  angle = (now->tm_sec - 15) * -0.1047;
  x = s_len * cos(angle);
  y = s_len * sin(angle);
  XDrawLine(display, w, gc, 50, 50, 50 + x, 50 - y);
}

void usage(char *title)
{
  fprintf(stderr, "Usage: %s [-face colour] [-hands colour]\n"
          "[-digits colour] [-image image] [-imagex x]\n"
          "[-imagey y] [-font font] [-drad rad] [-hlen len]\n"
          "[-mlen len] [-slen len] [-nodigits] [-top] [-wm]\n"
          "[-display display] [-geometry geometry]\n\n", title);

  fprintf(stderr, "-face colour (Choose a colour for the clock face.)\n"
          "-hands colour (Choose a colour for the hands.)\n"
          "-digits colour (Choose a colour for the digits.)\n"
          "-image image (Choose an xpm image for the face.)\n"
          "-imagex x (Offset the face image by x pixels to the right.)\n"
          "-imagey y (Offset the face image by y pixels down.)\n"
          "-font font (Font for digits.)\n"
          "-drad rad (Radius of digits.)\n"
          "-hlen len (Length of hour hand.)\n"
          "-mlen len (Length of minute hand.)\n"
          "-slen len (Length of second hand.)\n"
          "-nodigits (Don't draw digits on clock.)\n"
          "-top (Keep clock always on top of other windows.)\n"
          "-wm (Let the window manager supply decorations.)\n"
          "-display display (Specify the display to use.)\n"
          "-geometry geometry (Define the initial window geometry.)\n");

  exit(EXIT_FAILURE);
}
